#include <iostream>
#include "netcdf.h"

int main(void)
{
int ret, ncid;

if ((ret = nc_create("foo.nc", NC_CLOBBER, &ncid)))
  return ret;

if ((ret = nc_close(ncid)))
  return ret;

return EXIT_SUCCESS;
}
