; Autogenerated by the Meson build system.
; Do not edit, your changes will be lost.

%define bool

%undef false

; ultimate question of life, the universe, and everything
%define int 42

; This is
; a multiline
; description
%define str "hello world!"

%define unquoted float

