#!/usr/bin/env python3

import sys
import os

assert os.environ['MESONTESTING'] == 'picklerror'
assert os.environ['TEST_LIST_FLATTENING'] == '1'

with open(sys.argv[1]) as f:
    if f.read() != 'contents\n':
        sys.exit(1)
