#include <windows.h>

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  // avoid unused argument error while matching template
    ((void)hinstDLL);
    ((void)fdwReason);
    ((void)lpvReserved);
  return TRUE;
}
