#!/usr/bin/env python3

import sys

if len(sys.argv) != 3:
    print(sys.argv[0], '<input>', '<output>')

inf = sys.argv[1]
outf = sys.argv[2]

with open(outf, 'wb') as o:
    with open(inf, 'rb') as i:
        o.write(i.read())
